#
#
# Many Jobs Manager
#
# 11/17/2009 Hideki Fujioka
# 05/05/2011 Last Modified by Hideki Fujioka
#
import os
import sys
import commands
import string
import time
import fcntl
import getpass
#
from log_file import *
#
#
class ManyJobsManager:
    """ Many Jobs Manager """
    def __init__(self,fileName):
        """ lock DB and read it """
        self.manyjobsDB = os.path.abspath(fileName)
        self.localhost = ""
        self.hostName=[]
        self.ssh = "ssh" # default secure shell
        self.scp = "scp" # default secure Copy
        self.hostAddress=[]
        self.jobSubmitLauncher=[]
        self.maxJobs=[]
        #
        self.taskName=[]
        self.taskFile=[]
        self.taskDep=[]
        self.taskStatus=[]
        #
        self.lockFile = "/tmp/" + getpass.getuser()
        self.lockFile += os.path.basename(os.path.dirname(self.manyjobsDB)) 
        self.lockFile += os.path.basename(self.manyjobsDB) + ".lock"
        #
        #### LOCK THE LOCK FILE ####
        self.lockFileDescriptor = open(self.lockFile,"w+")
        fcntl.flock(self.lockFileDescriptor.fileno(),fcntl.LOCK_EX)
        #
        # Read DB
        inp = open(self.manyjobsDB,'r')
        #
        self.lines = inp.readlines()
        inp.close()
        #
        for line in self.lines:
            # skip comments
            if (line[0] == '#'):
                continue
            entry = line.split()
            if (len(entry) == 0):
                continue
            # get localhost
            if (entry[0] == 'LOCALHOST'):
                self.localhost = entry[1]
            # get Cluster List
            if (entry[0] == 'CLUSTER'):            
                self.hostName.append(entry[1])
                self.hostAddress.append(entry[2])
                self.jobSubmitLauncher.append(entry[3])
                self.maxJobs.append(entry[4])
            # get Tasks to do
            if ((entry[0] == 'TASK') or (entry[0] == 'JOB')):            
                self.taskName.append(entry[1])
                self.taskFile.append(entry[2])
                self.taskDep.append(entry[3])
                if (len(entry) < 5):
                    self.taskStatus.append("Queued")
                else:
                    self.taskStatus.append(entry[4])
            # user secure shell
            if (entry[0] == 'SSH'):
                self.ssh =""
                for i in range(len(entry)):
                    if (i == 0):
                        continue
                    self.ssh += entry[i] + ' '
            # user secure copy
            if (entry[0] == 'SCP'):
                self.scp =""
                for i in range(len(entry)):
                    if (i == 0):
                        continue
                    self.scp += entry[i] + ' '
        #
    def finalize(self):
        """ update DB and unlock DB """
        #
        # update DB
        for i in range(len(self.lines)):
            line = self.lines[i]
            # skip comments
            if (line[0] == '#'):
                continue
            entry = line.split()
            if (len(entry) == 0):
                continue
            # update localhost
            if (entry[0] == 'LOCALHOST'):
                self.lines[i] = "LOCALHOST %s\n" % (self.localhost)
            # update Cluster List
            if (entry[0] == 'CLUSTER'):            
                for j in range(len(self.hostName)):
                    if (entry[1] == self.hostName[j]):
                        self.lines[i] = "CLUSTER %s %s %s %s\n" % (self.hostName[j],self.hostAddress[j],self.jobSubmitLauncher[j],self.maxJobs[j])
                        break
            # update Tasks to do
            if (entry[0] == 'TASK') or (entry[0] == 'JOB'):            
                for j in range(len(self.taskName)):
                    if (entry[1] == self.taskName[j]):
                        self.lines[i] = "TASK %s %s %s %s\n" % (self.taskName[j],self.taskFile[j],self.taskDep[j],self.taskStatus[j])
                        break
        #
        # write to DB file
        outp = open(self.manyjobsDB,'wt')        
        for line in self.lines:
            outp.write(line)
        outp.close()
        #
        #### UNLOCK THE LOCK FILE ###
        fcntl.flock(self.lockFileDescriptor.fileno(),fcntl.LOCK_UN)
        self.lockFileDescriptor.close()
    #
    def clearDataBase(self):
        """ initialze all tasks """
        for i in range(len(self.taskStatus)):
            self.taskStatus[i] = ""
    #
    def getLocalHost(self):
        """ return localhost """
        return  self.localhost
    #
    #
    def getClusterList(self):
        """ returns cluster information """
        return [self.hostName,self.hostAddress,self.jobSubmitLauncher,self.maxJobs]
    #
    #
    def getClusterAddressByName(self,cname):
        """ returns Address for a specified name """
        for i in range(len(self.hostName)):
            if (self.hostName[i] == cname):
                return self.hostAddress[i]
        #
        return None
    #
    #
    def getJobSubmitLauncherByName(self,cname):
        """ returns Job Submit-Launcher for a specified name """
        for i in range(len(self.hostName)):
            if (self.hostName[i] == cname):
                return self.jobSubmitLauncher[i]
        #
        return None
    #
    #
    def assignNewTask(self,clusterName,jobsID):
        """ returns a new task information and change the status to running """
        for count in range(len(self.taskName)):
            if (self.taskStatus[count] == "Queued"):
                if (self.taskDep[count] == "None"):
                    # no depending task. just run it
                    self.taskStatus[count] = "Run"
                    return [self.taskName[count],self.taskFile[count]]
                #
                # This task has a depending task
                for count2 in range(len(self.taskName)):
                    if (self.taskName[count2] == self.taskDep[count]):    
                        if (self.taskStatus[count2] == "Done"):            
                            # depengin task has finished, so run it
                            self.taskStatus[count] = "Run"
                            return [self.taskName[count],self.taskFile[count]]                            
                        #
                        # The depending task hasn't finished yet
                        #msg = self.taskName[count] + " waits until " + self.taskName[count2] + " finishes"
                        #log_file(os.path.dirname(self.manytasksDB),msg)
                        break
                #
                if (self.taskName[count2] != self.taskDep[count]): 
                    # Something wrong: Depensding task not found
                    msg = self.taskName[count] + " depends on " + self.taskName[count2] + " but not found"
                    log_file(os.path.dirname(self.manyjobsDB),msg)
        #
        # No more task
        return [None,None]                
    #
    #
    def reportFinishedTask(self,taskName,stat):
        """ report finished task to store DB """
        for count in range(len(self.taskName)):
            if (self.taskName[count] == taskName):
                break
        #
        if (self.taskName[count] != taskName):
            # No task found
            return 
        #
        self.taskStatus[count] = "Done"
        if (stat == "ERROR"):
            self.taskStatus[count] = "ERROR"
        if (stat == "ERROR_CLUSTER"):
            self.taskStatus[count] = "ERROR_CLUSTER"
        if (stat == "ERROR_MANYJOBS"):
            self.taskStatus[count] = "ERROR_MANYJOBS"
        if (stat == "KILLED"):
            self.taskStatus[count] = "KILLED"
    #
    #
    def ifThingsToDo(self):
        """ return Ture if still things to do """
        for count in range(len(self.taskName)):
            if (self.taskStatus[count] == "Queued"):
                if (self.taskDep[count] == "None"):
                    return True
                #
                for count2 in range(len(self.taskName)):
                    if (self.taskName[count2] == self.taskDep[count]):    
                        if (self.taskStatus[count2] == "Done"):
                            return True
                    #
        #
        # No more task
        return False
        #
    #
    #
    def getSecureShell(self):
        """ return secure shell command, ssh is default"""
        return self.ssh
    #
    def getSecureCopy(self):
        """ return secure copy command, scp is default"""
        return self.scp
#       
#
#
#
    def numberOfTasksToRun(self):
        """ returns the number of tasks ready to run """
        ntask = 0
        for count in range(len(self.taskName)):
            if (self.taskStatus[count] == "Queued"):
                if (self.taskDep[count] == "None"):
                    # no depending task. its ready
                    ntask += 1
                #
                # This task has a depending task
                for count2 in range(len(self.taskName)):
                    if (self.taskName[count2] == self.taskDep[count]):    
                        if (self.taskStatus[count2] == "Done"):            
                            # depengin task has finished, so its readly
                            ntask += 1
                        break
                #
        return ntask
    #
    #
