#!/usr/bin/env python
#
# Job Agent 
# 08/04/2011 Created by Hideki Fujioka
# 08/12/2011 Modified by Hideki Fujioka
#
import sys
import string
import os
import commands
import time
import subprocess
#
from log_file import *
from pull_task import *
from finalize_job_remote import *
#
dirname = os.path.dirname(os.path.abspath(sys.argv[0]))
batch = os.getenv('MANYJOBS_BATCH')
remoteusrbinDir = os.path.dirname(dirname) + "/remoteusrbin"
#
clusterName = os.getenv('MANYJOBS_CLUSTER_NAME')
masterHost = os.getenv('MANYJOBS_MASTER_HOST')
masterDir = os.getenv('MANYJOBS_MASTER_DIR')
scriptDir = os.getenv('MANYJOBS_SCRIPT_DIR')
logDir = os.getenv('MANYJOBS_LOG_DIR')
masterJobsDB = os.getenv('MANYJOBS_MASTER_JOBS_DB')
mpi_run = os.getenv('MANYJOBS_MPI')
headnode = os.getenv('MANYJOBS_HEADNODE')
sshcommand = os.getenv('MANYJOBS_SSHCOMMAND')
#
#
total_cpus = 0
if (batch == 'pbs'):
    pbs_node_file = os.getenv('PBS_NODEFILE')
    jobID = os.getenv('PBS_JOBID')
    if pbs_node_file == None:
        print "NODE FILE NOT FOUND\n"
        sys.exit(-1)
    os.putenv('MANYJOBS_JOBID', jobID)
    # read node file
    f = open(pbs_node_file)
    nodes = f.readlines()
    f.close()
    # count
    node_dict = {}
    for node in nodes:
        node_dict[node] = nodes.count(node)
    total_cpus = len(nodes)
    #
    # print    
    msg = "Num of CPUs= %s" % total_cpus
    log_file(logDir,msg)
    for node in node_dict.keys():
    	msg = node.strip('\n') + ":ppn=" + str(node_dict[node])
        log_file(logDir,msg)
    #
elif (batch == 'sge'):
    pe_node_file = os.getenv('PE_HOSTFILE')
    jobID = os.getenv('JOB_ID')
    if pe_node_file == None:
        print "NODE FILE NOT FOUND\n"
        sys.exit(-1)
    os.putenv('MANYJOBS_JOBID', jobID)
    # read node file
    f = open(pe_node_file)
    nodes = f.readlines()
    f.close()
    # count
    for node in nodes:
    	#print "%s\n" % node
        ent = node.split()
        total_cpus = total_cpus + int(ent[1])
    # print    
    msg ="Num of CPUs= %s" % total_cpus
    log_file(logDir,msg)
else:
    msg = "BATCH SYSTEM %s IS NOT SUPPORTED\n" % (batch)
    log_file(logDir,msg)
    sys.exit(-1)
#
#
# BIG_JOB
numOfSubjobs = 1
minNumOfSubjobs = 1
if ('None' != os.getenv('MANYJOBS_NUMOFSUBJOBS')):
    try:
        numOfSubjobs = int(os.getenv('MANYJOBS_NUMOFSUBJOBS'))
    except:
        numOfSubjobs =  1
#
if ('None' != os.getenv('MANYJOBS_MINNUMOFSUBJOBS')):
    try:
        minMumOfSubjobs = int(os.getenv('MANYJOBS_MINNUMOFSUBJOBS'))
    except:
        minNumOfSubjobs = numOfSubjobs
#       
#
# Dvide CPUs into numOfSubjobs
myCPUs = []
for id in range(numOfSubjobs):
    myCPUs.append(total_cpus / numOfSubjobs) 
    if (total_cpus % numOfSubjobs != 0):
        if (id < total_cpus % numOfSubjobs):
            myCPUs[id] += 1
#
# Allocate CPUs 
idx = 0
myNodes = []
for id in range(numOfSubjobs):
    myNodes.append([])
    for n in range(myCPUs[id]):
        myNodes[id].append(nodes[idx])
        idx += 1
#        
# Create Nodes Files
nodeFileName = []
for id in range(numOfSubjobs):
    nodeFileName.append(logDir + '/' + jobID + "%04d.nod" % (id))
    outp = open(nodeFileName[id],'wt')        
    for node in myNodes[id]:
        outp.write(node)
    outp.close()
#
#
# Pull Task(s)
[numOfSubjobs, taskNames] = pull_task(jobID,numOfSubjobs,minNumOfSubjobs)
#
#
# Run SubJob(s)
proc = []
stat = []
if (numOfSubjobs >= 2):
    #SSH to RUN GET ENV & RUN
    for id in range(numOfSubjobs):
        subjob_master_node = (myNodes[id][0].split('\n'))[0]
        commandToLaunchSubJobs = sshcommand + " " + subjob_master_node + " "
        commandToLaunchSubJobs += dirname + "/run_subjob.py "
        commandToLaunchSubJobs += logDir + " "
        commandToLaunchSubJobs += jobID + " "
        commandToLaunchSubJobs += str(id)
        
        log_file(logDir,commandToLaunchSubJobs)
        p = subprocess.Popen(commandToLaunchSubJobs,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        proc.append(p)
        stat.append('RUN')
        msg = "A SubJob runs on " + subjob_master_node
        log_file(logDir,msg)
#
if (numOfSubjobs == 1):
    # run
    commandToLaunchSubJobs = dirname + "/run_subjob.py "
    commandToLaunchSubJobs += logDir + " "
    commandToLaunchSubJobs += jobID + " "
    commandToLaunchSubJobs += str(id)
    log_file(logDir,commandToLaunchSubJobs)
    p = subprocess.Popen(commandToLaunchSubJobs,
                         shell=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    proc.append(p)
    stat.append('RUN')
    msg = "A Task runs on here"
    log_file(logDir,msg)    
#
#
status = 'None'
if (numOfSubjobs >= 1):
    # Wait Until All subjobs done
    while(True):
        runningtasks = 0
        for id in range(numOfSubjobs):
            if (stat[id] == 'RUN'):
                runningtasks += 1
                if(None != proc[id].poll()):
                    # get status
                    stat[id] = 'None'
                    comm = proc[id].communicate()
                    sout = comm[0]
                    serr = comm[1]
                    print "*********TASK_ID=" + str(id) + "********\n"
                    print sout
                    print "**************************\n"
                    print >>sys.stderr,"*********TASK_ID=" + str(id) + "********\n"
                    print >>sys.stderr,serr
                    print >>sys.stderr,"**************************\n"
                    lines = sout.split('\n')
                    for line in lines:
                        ent = line.split('=')
                        if (ent[0].strip() == "MANYJOBS_STATUS"):
                            stat[id] = ent[1].strip()
                    #
            if (stat[id] == 'ERROR'):
                print >>sys.stderr,str(id) + " aborts with ERROR" 
            if (stat[id] == 'ERROR_CLUSTER') or (stat[id] == 'ERROR_MANYJOBS'): 
                status = stat[id]
                print >>sys.stderr,str(id) + " aborts with " + status 
        #
        #print str(runningtasks) + " running on " + jobID
        if (runningtasks == 0):
            print "All SubJobs completed"
            break
        if (status != 'None'):
            print "A SubJob aborts with errors"
            break
        # wait 
        time.sleep(60)
    #while()
    
    # If exit with errors, kill all tasks
    if (status == 'ERROR_CLUSTER') or (status == 'ERROR_MANYJOBS'):     
        for id in range(numOfSubjobs):
            if(None == proc[id].poll()):        
                # KILL
                proc[id].kill()
                stat[id] = "Killed"
    #
    #
    # Report End Tasks Status
    end_task_remote(jobID,numOfSubjobs,taskNames,stat)
    #
#
#
# FINILIZE THE JOB
finalize_job_remote(jobID,numOfSubjobs,status)


